package com.spotify.helios.servicescommon.coordination;

import com.spotify.helios.ZooKeeperStandaloneServerManager;
import com.spotify.helios.servicescommon.DefaultZooKeeperClient;

import org.apache.commons.io.FileUtils;
import org.apache.curator.utils.ZKPaths;
import org.apache.zookeeper.KeeperException;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Arrays;
import java.util.concurrent.Callable;

import static com.spotify.helios.Polling.await;
import static com.spotify.helios.servicescommon.coordination.ZooKeeperModelReporter.noop;
import static java.util.concurrent.TimeUnit.SECONDS;
import static org.junit.Assert.assertArrayEquals;

public class ZooKeeperUpdatingPersistentDirectoryTest {

  private static final String PARENT_PATH = "/foobar";

  private Path stateFile;
  private Path backupDir;

  private static final byte[] BAR1_DATA = "bar1".getBytes();
  private static final byte[] BAR2_DATA = "bar2".getBytes();
  private static final byte[] BAR3_DATA = "bar2".getBytes();
  private static final String FOO_NODE = "foo";
  private static final String BAZ_NODE = "baz";
  private static final String FOO_PATH = ZKPaths.makePath(PARENT_PATH, FOO_NODE);
  private static final String BAZ_PATH = ZKPaths.makePath(PARENT_PATH, BAZ_NODE);

  private ZooKeeperStandaloneServerManager zk = new ZooKeeperStandaloneServerManager();

  private ZooKeeperUpdatingPersistentDirectory sut;

  @Before
  public void setUp() throws Exception {
    backupDir = Files.createTempDirectory("helios-zk-updating-persistent-dir-test-backup-");
    stateFile = Files.createTempFile("helios-zk-updating-persistent-dir-test-", "");
    zk.curator().newNamespaceAwareEnsurePath(PARENT_PATH).ensure(zk.curator().getZookeeperClient());
    setupDirectory();
  }

  private void setupDirectory() throws IOException {
    final DefaultZooKeeperClient client = new DefaultZooKeeperClient(zk.curator());
    final ZooKeeperClientProvider provider = new ZooKeeperClientProvider(client, noop());
    sut = ZooKeeperUpdatingPersistentDirectory.create("test", provider, stateFile, PARENT_PATH);
    sut.startAsync();
  }

  @After
  public void tearDown() throws Exception {
    sut.stopAsync().awaitTerminated();
    zk.close();
    FileUtils.deleteQuietly(stateFile.toFile());
    FileUtils.deleteQuietly(backupDir.toFile());
  }

  @Test
  public void verifyCreatesAndRemovesNode() throws Exception {
    sut.map().put(FOO_NODE, BAR1_DATA);
    final byte[] remote = awaitNode(FOO_PATH);
    assertArrayEquals(BAR1_DATA, remote);
    sut.map().remove(FOO_NODE);
    awaitNoNode(FOO_PATH);
  }

  @Test
  public void verifyUpdatesDifferingNode() throws Exception {
    zk.curator().create().forPath(FOO_PATH, "old".getBytes());
    sut.map().put(FOO_NODE, BAR1_DATA);
    awaitNodeWithData(FOO_PATH, BAR1_DATA);
  }

  @Test
  public void verifyRemovesUndesiredNode() throws Exception {
    zk.curator().create().forPath(FOO_PATH, BAR1_DATA);
    zk.stop();
    zk.start();
    awaitNoNode(FOO_PATH);
  }

  @Test
  public void verifyRecoversFromBackupRestoreOnline() throws Exception {
    // Create backup
    zk.curator().create().forPath("/version", "1".getBytes());
    sut.map().put(FOO_NODE, BAR1_DATA);
    awaitNodeWithData(FOO_PATH, BAR1_DATA);
    zk.backup(backupDir);

    // Write data after backup
    zk.curator().setData().forPath("/version", "2".getBytes());
    sut.map().put(FOO_NODE, BAR2_DATA);
    sut.map().put(BAZ_NODE, BAR3_DATA);
    awaitNodeWithData(FOO_PATH, BAR2_DATA);
    awaitNodeWithData(BAZ_PATH, BAR3_DATA);

    // Restore backup
    zk.stop();
    zk.restore(backupDir);
    zk.start();
    assertArrayEquals("1".getBytes(), zk.curator().getData().forPath("/version"));

    // Verify that latest data is pushed
    awaitNodeWithData(FOO_PATH, BAR2_DATA);
    awaitNodeWithData(BAZ_PATH, BAR3_DATA);
  }

  @Test
  public void verifyRecoversFromBackupRestoreOffline() throws Exception {
    // Create backup
    zk.curator().create().forPath("/version", "1".getBytes());
    sut.map().put(FOO_NODE, BAR1_DATA);
    awaitNodeWithData(FOO_PATH, BAR1_DATA);
    zk.backup(backupDir);

    // Write data after backup
    zk.curator().setData().forPath("/version", "2".getBytes());
    sut.map().put(FOO_NODE, BAR2_DATA);
    sut.map().put(BAZ_NODE, BAR3_DATA);
    awaitNodeWithData(FOO_PATH, BAR2_DATA);
    awaitNodeWithData(BAZ_PATH, BAR3_DATA);

    // Stop persistent directory
    sut.stopAsync().awaitTerminated();

    // Restore backup
    zk.stop();
    zk.restore(backupDir);
    zk.start();
    assertArrayEquals("1".getBytes(), zk.curator().getData().forPath("/version"));

    // Start new persistent directory
    setupDirectory();

    // Verify that latest data is pushed
    awaitNodeWithData(FOO_PATH, BAR2_DATA);
    awaitNodeWithData(BAZ_PATH, BAR3_DATA);
  }

  private byte[] awaitNode(final String path) throws Exception {
    return await(30, SECONDS, new Callable<byte[]>() {
      @Override
      public byte[] call() throws Exception {
        try {
          return zk.curator().getData().forPath(path);
        } catch (KeeperException.NoNodeException e) {
          return null;
        }
      }
    });
  }

  private void awaitNodeWithData(final String path, final byte[] data) throws Exception {
    await(30, SECONDS, new Callable<Boolean>() {
      @Override
      public Boolean call() throws Exception {
        try {
          final byte[] remote = zk.curator().getData().forPath(path);
          return Arrays.equals(data, remote) ? true : null;
        } catch (KeeperException.NoNodeException e) {
          return null;
        }
      }
    });
  }

  private void awaitNoNode(final String path) throws Exception {
    await(30, SECONDS, new Callable<Boolean>() {
      @Override
      public Boolean call() throws Exception {
        try {
          zk.curator().getData().forPath(path);
          return null;
        } catch (KeeperException.NoNodeException e) {
          return true;
        }
      }
    });
  }
}
